#--
# *** This code is copyright 2005 by Gavin Kistner
# *** It is covered under the license viewable at http://phrogz.net/JS/_ReuseLicense.txt
# *** Reuse or modification is free provided you abide by the terms of that license.
# *** (Including the first two lines above in your source code usually satisfies the conditions.)
#++
# This file covers the OWLScribble class, and extensions to the String
# class needed by it.
# Please see the documentation on those classes for more information.
#
# See the link:../markup.html file for details on the OWLScribble markup.
#
# Author::     Gavin Kistner  (mailto:!@phrogz.net)
# Copyright::  Copyright (c)2005 Gavin Kistner
# License::    See http://Phrogz.net/JS/_ReuseLicense.txt for details
# Version::    0.8 (2005-July-4)
# Full Code::  link:../OWLScribble.rb
#
# == TODO
# * Speed optimizations, once it's feature-complete
# * Process indented paragraphs like lists, just to calculate minimal depth from many spaces
# * Inline commands for <tt><br></tt>? Some auto-br inclusion?
# * More unit tests

require 'TagTreeScanner.rb'

# OWLScribble converts a specific set of text markup into HTML.
# (The syntax used in the markup is a knockoff of the markup used by
# OpenWiki, so the 'OWL' in OWLScribble means "OpenWiki-like".)
#
# The tree of elements created during parsing is made available, along
# with collections for specific types of elements, such as a list of all
# wiki-commands or wiki page links. This lets you modify the element tree
# (e.g. perform a database lookup on pages and replace the links with 
# html anchors to your URL) after parsing, but before creating the HTML
# string.
#
# = Usage
# <b>Basic Example:</b>
#  owl = OWLScribble.new( the_string_to_parse )
#  puts owl.to_html
#
# <b>Modifying the Tree:</b>
#  require 'uri'
#  owl.wiki_links.each{ |tag|
#    tag.tag_name = :a
#    page = tag.attributes[ :page ]
#    tag.attributes[ :href ] = "view?#{URI.escape(page)}"
#    tag.attributes.delete( :page )
#  }
#
#  owl.wiki_commands.each{ |tag|
#    command = tag.attributes[ :do ]
#    params  = tag.attributes[ :param ]
#    # ...
#  }
#
#  puts owl.to_html
#
# = Wiki Links
# OWLScribble supports three ways to specify a link to a specific 'page',
# indicated by name:
# * <b><tt>WikiWords</tt></b> <i>(aka CamelCase)</i> are assumed to be page links.
# * <b><tt>[WikiWord some other text]</tt></b> creates a link to a page with alternate text.
# * <b><tt>[[Some Page! Name]]</tt></b> allows link creation to pages whose names are not WikiWords.
# <i>(For more information on creating links, see the link:../markup.html file
# supplied with this library.)</i>
#
# Because OWLScribble cannot know how to create URLs for such items, it
# simply wraps the text to link in a special "wiki_link" tag, with
# a <tt>:page</tt> attribute holding the name of the page to be linked to.
#
# Instances of the OWLScribble class have a <tt>wiki_links</tt> property
# that provides you with an array of such links, so that you can iterate
# these links and manipulate them before generating HTML.
#
# The following example snippet shows how to mutate these non-HTML tags
# into HTML anchors with a custom url, leaving the text of the link
# untouched:
#
#  require 'uri'
#  owl.wiki_links.each{ |tag|
#    tag.tag_name = :a
#    page = tag.attributes[ :page ]
#    tag.attributes[ :href ] = "view?#{URI.escape(page)}"
#    tag.attributes.delete( :page )
#  }
#
# = Wiki Commands
# In addition to links to wiki pages, OWLScribble allows the user to
# specify three commands:
# * <tt>##TableOfContents##</tt> - include a table of contents here
# * <tt>##DEPRECATED##</tt> - mark this page for future deletion
# * <tt>##Include(PageName)##</tt> - include the contents of another page at this location
# * <tt>##RedirectTo(PageName)##</tt> - show the specified page instead
# 
# OWLScribble handles the first automatically, building a table of contents
# from the hierarchy of heading levels, and replacing any <tt>##TableOfContents##</tt>
# commands with copies of that TOC. Additionally, every section following
# a header tag is wrapped in a <tt><div class="section">...</div></tt> block,
# providing a nested hierarchy of sections in the page itself. (This is 
# primarily useful for applying CSS to sections, such as visual indentation.)
#
# For the other commands, it is up to you to handle them. Instances of the
# OWLScribble class provide a +wiki_commands+ property, which is an array
# of <tt>"wiki_command"</tt> tags. These tags have a <tt>:do</tt> attribute
# which is the string name of the command, and (for +Include+ and +RedirectTo+)
# a <tt>:param</tt> attribute.
#
# If they are not removed, these tags will show up in the HTML output.
# For example, this input text:
#  == Engineering ==
#  Here's the home page from engineering:
#  
#  ##Include(EngineeringHome)##
#
# will produce the following HTML:
#  <h2 id="Engineering">Engineering</h2>
#  <div class="section">
#  <p>Here's the home page from engineering:</p>
#  <wiki_command do="Include" param="EngineeringHome"></wiki_command>
#  </div>
#
# The presence of these tags should not affect the functioning or rendering
# of the produced HTML in any modern browser, but it will be syntactically invalid.
#
# = \OWLScribble Markup
# For details on the markup used by OWLScribble, please see the ../link:markup.html
# file.
class OWLScribble < TagTreeScanner
	attr_reader :list_items #:nodoc:
	attr_reader :headers #:nodoc:

	# An array of <tt><wiki_command></tt> tags repesenting directives to the wiki. <i>(See "Wiki Commands" above)</i>
	attr_reader :wiki_commands

	# An array of <tt><wiki_link></tt> tags repesenting links to wiki pages. <i>(See "Wiki Links" above)</i>
	attr_reader :wiki_links

	# A nested unordered list representing the hierarchy of the document, with links to the sections. <i>(See "Table of Contents" above)</i>
	attr_reader :table_of_contents
	
	@root_factory = TagFactory.new( :root,
	  :allowed_genre => :block,
	  :allows_text => false )

	# Letters and numbers cannot be mixed with whitespace due to http:// as opener.
	@text_match = /[a-z0-9]+|[ \t]+|./mi

	@tag_genres[ :block ] = [ ]
	@tag_genres[ :table ] = [ ]
	@tag_genres[ :td ] = [ ]
	@tag_genres[ :deflist ] = [ ]
	@tag_genres[ :inline ] = [ ]

	# == My Heading ==  ->  <h2>My Heading</h2>
	@tag_genres[ :block ] << TagFactory.new( :heading,
		:open_match => /(={1,6}) +(.+) +\1[ \t]*\n/, :open_requires_bol => true,
		:setup => lambda{ |tag, ss, owl|
			depth = ss[1].length
			tag.tag_name = "h#{depth}"
			tag.info[ :header_depth ] = depth
			tag << ss[2]
			owl.headers << tag
		},
		:allows_text => true,
		:autoclose => true
	)

	#   * This is a bullet  ->  <li>This is a bullet</li>
	# (The private #reparent_lists method adds <ul>...</ul>)
	@tag_genres[ :block ] << TagFactory.new( :li,
		:open_match => /([ \t]+)\* /, :open_requires_bol => true,
		:close_match => /\n/,
		:setup => lambda{ |tag, ss, owl|
			tag.info[ :level ] = ss[1].length
			tag.info[ :list_type ] = :ul
			owl.list_items << tag
		},
		:allows_text => true,
		:allowed_genre => :inline
	)

	#   # Make functional  ->  <li>Make functional</li>
	# (The private #reparent_lists method adds <ol>...</ol>)
	@tag_genres[ :block ] << TagFactory.new( :li,
		:open_match => /([ \t]+)# /, :open_requires_bol => true,
		:close_match => /\n/,
		:setup => lambda{ |tag, ss, owl|
			tag.info[ :level ] =  ss[1].length
			tag.info[ :list_type ] = :ol
			owl.list_items << tag
		},
		:allows_text => true,
		:allowed_genre => :inline
	)

	#   2. Add good comments  ->  <li>Add good comments</li>
	# (The private #reparent_lists method adds <ol>...</ol>)
	@tag_genres[ :block ] << TagFactory.new( :li,
		:open_match => /([ \t]+)\d+\. /, :open_requires_bol => true,
		:close_match => /\n/,
		:setup => lambda{ |tag, ss, owl|
			tag.info[ :level ] =  ss[1].length
			tag.info[ :list_type ] = :ol
			owl.list_items << tag
		},
		:allows_text => true,
		:allowed_genre => :inline
	)

	#   a. alpha 1  ->  <li type="a">alpha 1</li>
	#   A. ALPHA 2  ->  <li type="A">ALPHA 2</li>
	#   i. roman 3  ->  <li type="i">roman 3</li>
	#   I. ROMAN 4  ->  <li type="I">ROMAN 4</li>
	# (The private #reparent_lists method adds <ol>...</ol>)
	@tag_genres[ :block ] << TagFactory.new( :li,
		:open_match => /([ \t]+)([aAiI])\. /, :open_requires_bol => true,
		:close_match => /\n/,
		:setup => lambda{ |tag, ss, owl|
			tag.attributes[ :type ] = ss[2]
			tag.info[ :level ] = ss[1].length
			tag.info[ :list_type ] = :ol
			owl.list_items << tag
		},
		:allows_text => true,
		:allowed_genre => :inline
	)

	#   ; Semantic : relating to meaning  ->  <dl><dt>Semantic</dt><dd>relating to meaning</dd></dl>
	# ( requires the :dt and :dd matches in the :deflist genre )
	@tag_genres[ :block ] << TagFactory.new( :dl,
		:open_match => /(?=[ \t]+; .+ : )/, :open_requires_bol => true,
		:close_match => /(?![ \t]+; .+ : )/, :close_requires_bol => true,
		:allowed_genre => :deflist
	)

	# {{{
	# This is a multi-line
	# code block
	# }}}
	# -> <pre>This is a multi-line
	#    code block</pre>
	@tag_genres[ :block ] << TagFactory.new( :pre,
		:open_match => /^([ \t]*)\{\{\{\n(.+?)\n\1\}\}\}\n/m, :open_requires_bol => true,
		:close_match => /\n/,
		:setup => lambda{ |tag, ss, owl| tag << ss[2].gsub( /^#{ss[1]}/, '' ) },
		:allows_text => true,
		:autoclose => true
	)

	# || Age || Sex ||
	# || 32  ||  M  ||
	# -> <table><tr><td>Age</td><td>Sex</td></tr><tr><td>32</td><td>M</td></tr></table>
	# ( requires the :tr factory in the :table genre, and the :td factory in the :td genre )
	@tag_genres[ :block ] << TagFactory.new( :table,
		:open_match => /(?=\|\|)/, :open_requires_bol => true,
		:close_match => /(?=[^|])/, :close_requires_bol => true,
		:allowed_genre => :table
	)

	#   : Indented Text!  ->  <p class="indent2">Indented Text!</p>
	@tag_genres[ :block ] << TagFactory.new( :p,
		:open_match => /((  )+) ?: /, :open_requires_bol => true,
		:close_match => /\n/,
		:setup => lambda{ |tag, ss, owl| tag.attributes[ :class ] = "indent#{ss[1].length/2}" },
		:allows_text => true,
		:allowed_genre => :inline
	)

	# ##TableOfContents##  ->  <wiki_command do="TableOfContents" />
	# ##IncludePage(FooBar)##  ->  <wiki_command do="IncludePage" param="FooBar" />	
	@tag_genres[ :block ] << TagFactory.new( :wiki_command,
		:open_match => /##(TableOfContents|Deprecated|Baleted|Include|RedirectTo)(?:\s*\(\s*(.+)\s*\))?##/, :open_requires_bol => true,
		:setup => lambda{ |tag, ss, owl|
			tag.attributes[ :do ] = ss[1][ /[a-z]+/i ]
			tag.attributes[ :param ] = ss[2] if ss[2]
			owl.wiki_commands << tag
		},
		:autoclose => true
	)

	# ----  ->  <hr />
	@tag_genres[ :block ] << TagFactory.new( :hr,
		:open_match => /-{4,} *\n/, :open_requires_bol => true,
		:autoclose => true
	)

	# All other lines that begin with non-whitespace start a paragraph
	@tag_genres[ :block ] << TagFactory.new( :p,
		:open_match => /(?=\S)/, :open_requires_bol => true,
		# Stop the paragraph for a double-newline or a line that looks like a list or header start
		:close_match => /\n(?:\n|(?=[ \t]+(?:[*aiAI#]|\d+\.) )|(?=(={1,6}) +.+ +\1[ \t]*\n))/,
		:allows_text => true,
		:allowed_genre => :inline
	)

	# All other lines that begin with whitespace and have non-whitespace
	# content later start a preformatted block
	@tag_genres[ :block ] << TagFactory.new( :pre,
		:open_match => /([ \t]+)([^\n]*\S.+?)^(?=\S)/m, :open_requires_bol => true,
		:setup => lambda{ |tag, ss, owl|
			tag << ss[2].gsub( /^#{ss[1]}/, '' )
		},
		:allows_text => true,
		:autoclose => true
	)


	# (see the :table factory in :block genre)
	@tag_genres[ :table ] << TagFactory.new( :tr,
		:open_match => /(?=\|\|)/, :open_requires_bol => true,
		:close_match => /\|\|[ \t]*\n/,
		:allowed_genre => :td
	)


	# (see the :table factory in :block genre)
	@tag_genres[ :td ] << TagFactory.new( :td,
		:open_match => /((?:\|\|)+)\s*/,
		:close_match => /(?=\s*\|\|)/,
		:setup => lambda{ |tag, ss, owl|
			colspan = ss[1].length / 2
			tag.attributes[ :colspan ] =  colspan unless colspan < 2
		},
		:allows_text => true,
		:allowed_genre => :inline
	)


	# (see the :dl factory in :block genre)
	@tag_genres[ :deflist ] << TagFactory.new( :dd,
		:open_match => /[ \t]+;\s+: /, :open_requires_bol => true,
		:close_match => /\n/,
		:allows_text => true,
		:allowed_genre => :inline
	)
	@tag_genres[ :deflist ] << TagFactory.new( :dt,
		:open_match => /[ \t]+; /, :open_requires_bol => true,
		:close_match => / : /,
		:allows_text => true,
		:allowed_genre => :inline
	)
	@tag_genres[ :deflist ] << TagFactory.new( :dd,
		:open_match => /(?=.)/,
		:close_match => /\n/,
		:allows_text => true,
		:allowed_genre => :inline
	)


	# This is **bold text**  ->  This is <b>bold text</b>
	@tag_genres[ :inline ] << TagFactory.new( :b,
		# Close up if we see a newline to prevent a typo from propagating too far
		:open_match => /\*\*/, :close_match => /\*\*|(?=\n)/,
		:allows_text => true,
		:allowed_genre => :inline
	)

	# This is //italic text//  ->  This is <i>italic text</i>
	@tag_genres[ :inline ] << TagFactory.new( :i,
		# Close up if we see a newline to prevent a typo from propagating too far
		:open_match => /\/\//, :close_match => /\/\/|(?=\n)/,
		:allows_text => true,
		:allowed_genre => :inline
	)

	# This is @@literal text@@  ->  This is <tt>literal text</tt>
	@tag_genres[ :inline ] << TagFactory.new( :tt,
		:open_match => /@@([^\n]+?)@@/,
		:setup => lambda{ |tag, ss, owl| tag << ss[1] },
		:allows_text => true,
		:autoclose => true
	)

	# This is {{{also literal text}}}  ->  This is <tt>also literal text</tt>
	@tag_genres[ :inline ] << TagFactory.new( :tt,
		:open_match => /\{\{\{([^\n]+?)\}\}\}/,
		:setup => lambda{ |tag, ss, owl| tag << ss[1] },
		:allows_text => true,
		:autoclose => true
	)

	# hTTp://www.google.com  ->  <a href="hTTp://www.google.com">www.google.com</a>
	# Require capital 'T' as a lame way to fight spammers
	@tag_genres[ :inline ] << TagFactory.new( :a,
		:open_match => /(?:hTTp|fTp|hTTps):\/\/(\S+)/,
		:setup => lambda{ |tag, ss, owl|
			tag.attributes[ :href ] = ss[0]
			tag << ss[1]
		},
		:allows_text => true,
		:autoclose => true
	)

	# http://www.google.com  ->  http://www.google.com
	# ensure that double-slash from non-link doesn't start italics
	@tag_genres[ :inline ] << TagFactory.new( :span,
		:open_match => /(?:http|ftp|https):\/\/(\S+)/i,
		:setup => lambda{ |tag, ss, owl|
			tag << ss[0]
		},
		:allows_text => true,
		:autoclose => true
	)

	# [hTTp://www.google.com Google is cool!] ->  <a href="hTTp://www.google.com">Google is cool!</a>
	# Require capital 'T' as a lame way to fight spammers
	@tag_genres[ :inline ] << TagFactory.new( :a,
		:open_match => /\[((?:hTTp|fTp|hTTps):\/\/\S+) +([^\]]+)\]/,
		:setup => lambda{ |tag, ss, owl|
			tag.attributes[ :href ] = ss[1]
			tag << ss[2]
		},
		:allows_text => true,
		:autoclose => true
	)

	# See the WhatWeShouldDo page  ->  See the <wiki_link page="WhatWeShouldDo">What We Should Do</wiki_link> page
	# OWLScribble#wiki_links gives you a collection; use Tag#replace_with to swap in what you want before #to_html.
	@tag_genres[ :inline ] << TagFactory.new( :wiki_link,
		:open_match => /[A-Z]{2,}[a-z][a-zA-Z]*|[A-Z][a-z]+[A-Z][a-zA-Z]*/,
		:setup => lambda{ |tag, ss, owl|
			tag.attributes[ :page ] = ss[0]
			tag << ss[0].dewikiword
			owl.wiki_links << tag
		},
		:allows_text => true,
		:autoclose => true
	)

	# See the [[Crazy! Go Nuts!]] page  ->  See the <wiki_link page="Crazy! Go Nuts!">Crazy! Go Nuts!</wiki_link> page
	# OWLScribble#wiki_links gives you a collection; use Tag#replace_with to swap in what you want before #to_html.
	@tag_genres[ :inline ] << TagFactory.new( :wiki_link,
		:open_match => /\[\[(.{2,}?)\]\]/,
		:setup => lambda{ |tag, ss, owl|
			tag.attributes[ :page ] = ss[1]
			tag << ss[1]
			owl.wiki_links << tag
		},
		:allows_text => true,
		:autoclose => true
	)

	# See the [WikiInstructions help page] for more  ->  See the <wiki_link page="WikiInstructions">help page</wiki_link> for more
	@tag_genres[ :inline ] << TagFactory.new( :wiki_link,
		:open_match => /\[([A-Z]{2,}[a-z][a-zA-Z]*|[A-Z][a-z]+[A-Z][a-zA-Z]*) ([^\]]+)\]/,
		:setup => lambda{ |tag, ss, owl|
			tag.attributes[ :page ] = ss[1]
			tag << ss[2]
			owl.wiki_links << tag
		},
		:allows_text => true,
		:autoclose => true
	)
	
	# ~NotALink  ->  <span>NotALink</span>
	@tag_genres[ :inline ] << TagFactory.new( :span,
		:open_match => /~([A-Z]{2,}[a-z][a-zA-Z]*|[A-Z][a-z]+[A-Z][a-zA-Z]*|\[\[([^\]\n]{2,}?)\]\]|\[(?:[A-Z]{2,}[a-z][a-zA-Z]*|[A-Z][a-z]+[A-Z][a-zA-Z]*) [^\]]+\])/,
		:setup => lambda{ |tag, ss, owl| tag << ss[1] },
		:allows_text => true,
		:autoclose => true
	)

	# This is --strike text--  ->  This is <strike>strike text</strike>
	@tag_genres[ :inline ] << TagFactory.new( :strike,
		# Close up if we see a newline to prevent a typo from propagating too far
		:open_match => /--/, :close_match => /--|(?=\n)/,
		:allows_text => true,
		:allowed_genre => :inline
	)

	# We should... !!finish this sentence!!  ->  We should... <span class="todo">TODO - finish this sentence</span>
	@tag_genres[ :inline ] << TagFactory.new( :span,
		:open_match => /!!([a-z].+?)!!/i,
		:setup => lambda{ |tag, ss, owl|
			tag << "TODO - #{ss[1]}"
			tag.attributes[ :class ] = 'todo'
		},
		:allows_text => true,
		:autoclose => true
	)

	# This is ^^superscript text^^  ->  This is <sup>superscript text</sup>
	@tag_genres[ :inline ] << TagFactory.new( :sup,
		# Close up if we see a newline to prevent a typo from propagating too far
		:open_match => /\^\^/, :close_match => /\^\^|(?=\n)/,
		:allows_text => true,
		:allowed_genre => :inline
	)

	# This is __subscript text__  ->  This is <sub>subscript text</sub>
	@tag_genres[ :inline ] << TagFactory.new( :sub,
		# Close up if we see a newline to prevent a typo from propagating too far
		:open_match => /__/, :close_match => /__|(?=\n)/,
		:allows_text => true,
		:allowed_genre => :inline
	)


	def initialize( owl_string )
		@list_items = []
		@headers = []
		@wiki_commands = []
		@wiki_links = []

		super

		reparent_lists( @list_items ) unless @list_items.empty?
		#TODO - process indented paragraphs like lists, just to calculate minimal depth from many spaces
		add_sections
		process_commands
		#process_links
	end

	private
		# Wrap <ul> and <ol> around tags appropriately
		# Extra work required to handle malformed nesting reasonably
		def reparent_lists( list_items )
			# Gather information about the tags before we munge the tree
			list_items.each_with_index{ |tag, i|
				tag.info[ :index ] = i
				tag.info[ :new_level ] = !tag.previous_sibling || ( tag.previous_sibling.tag_name != :li )
			}

			factories = {
				:ul => TagFactory.new( :ul ),
				:ol => TagFactory.new( :ol )
			}

			last_item, seen_at_level = nil
			list_items.dup.each{ |item|
				# cached for (premature) speed optimization
				info = item.info

				# reset history if we're in a new list
				if info[ :new_level ]
					seen_at_level = [ ]
					last_item = nil
				end

				# cached for (premature) speed optimization
				level = info[ :level ]

				# factory to create the <ul> or <ol>
				factory = factories[ item.info[ :list_type ] ]

				# add as child of last item, if I'm any deeper
				if last_item && level > last_item.info[ :level ]
					group_parent = last_item.append_child( factory.create )
					last_item.info[ :child_grouper ] = group_parent

				# at same level or higher than previous item
				else
					# sibling or parents (my level or higher)
					relatives = seen_at_level[ 1..level ]

					# find the most recent sibling/parent
					if relatives && ( parent_or_sib = relatives.compact.sort_by{ |tag| tag.info[ :index ] }.last )

						# same level as me? It's a sibling!
						if parent_or_sib.info[ :level ] == level
							group_parent = parent_or_sib.parent_tag

						# must be higher, thus it'll be my parent
						else
							group_parent = parent_or_sib.info[ :child_grouper ]
						end

					# I'm higher than anyone else (first item or malformed input)
					else
						group_parent = item.parent_tag.insert_before( factory.create, item )
					end

				end
				group_parent.append_child( item )

				seen_at_level[ level ] = item
				last_level = level
				last_item = item
			}
		end

		# Wrap <div class="section">...</div> around all headers
		# and create the table of contents
		def add_sections( )
			@table_of_contents = TagFactory.new( :div, :attributes => { :class => 'toc' } ).create

			section_factory = TagFactory.new( :div, :attributes => { :class => 'section' } )

			current_section = nil
			last_header = nil
			head_count = 0
			header_titles = Hash.new( 0 )
			header_at_depth = []
			relatives = nil

			# dup the array so that changes to the hierarchy
			# don't screw up the iteration
			@root.child_tags.dup.each{ |tag|
				# We assume that only headers have 'header_depth' information
				# because we're too lazy to test for h1-h6 tag_name
				if depth = tag.info[ :header_depth ]
					tag.info[ :head_count ] = head_count += 1

					# Create a TOC entry for this header
					header_title = tag.inner_text.gsub( /\W/, '' )
					header_id = header_title = tag.inner_text.gsub( /\W/, '' )
					hits = ( header_titles[ header_title ] += 1 )
					header_id += hits.to_s if hits > 1
					#header_id = "heading_#{head_count}"
					tag.attributes[ :id ] = header_id
					toc_item = @table_of_contents.append_child( Tag.new( :li ) )
					toc_item.info = { :level => depth, :list_type => :ul }
					anchor = toc_item.append_child( tag.dup )
					anchor.tag_name = :a
					anchor.attributes = { :href => "##{header_id}" }
					@table_of_contents.append_child( toc_item )

					# The wrapper div corresponding to this header
					section = tag.info[ :section ] = section_factory.create

					# if I'm any deeper than the last header, I'm a direct child
					if last_header && depth > last_header.info[ :header_depth ]
						section_parent = current_section

					# See if I can find a previous sibling or parent (my level or above)
					elsif ( relatives = header_at_depth[ 1..depth ] ) &&
					      ( parent_or_sib = relatives.compact.sort_by{ |header| header.info[ :head_count ] }.last )

						if parent_or_sib.info[ :header_depth ] == depth
							# same level as me? It's a sibling!
							section_parent = parent_or_sib.parent_tag
						else
							# must be higher, thus its section will be my parent
							section_parent = parent_or_sib.info[ :section ]
						end

					# I've got no basis for attachment, go to the 'top' level
					else
						section_parent = tag.parent_tag
					end
					section_parent.append_child( tag )
					section_parent.append_child( section )

					header_at_depth[ depth ] = tag
					current_section = section
					last_header = tag

				# we didn't find a header, so just shove the current
				# tag into the desired depth
				elsif current_section
					current_section.append_child( tag )
				end
			}

			reparent_lists( @table_of_contents.child_tags )
		end

		def process_commands( )
			@wiki_commands.each{ |command|
				case command.attributes[ :do ]
					when 'TableOfContents'
						command.replace_with( @table_of_contents.dup )
				end
			}
		end		
end

# Extensions to the built-in String class
class String
	# Returns a copy of the string with spaces inserted in the 'appropriate'
	# spots. e.g.
	#
	#  "WikiWord".dewikiword     #->  "Wiki Word"
	#  "OWLScribble".dewikiword  #->  "OWL Scribble"
	#  "ToXML".dewikiword        #->  "To XML"
	def dewikiword
		self.dup.dewikiword!
	end

	# Same as #dewikiword, but modifies the string in place
	def dewikiword!
		self.gsub!( /(.)([A-Z][a-z])/, '\\1 \\2' )
		self.gsub!( /([a-z])([A-Z])/, '\\1 \\2' )		
		self.gsub!( / (?:A|Of|The)(?= )/ ){ |txt| txt.downcase }
		self
	end
end
